/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  This program is free software; you can redistribute it and/or modify    |
 |  it under the terms of the GNU General Public License as published by    |
 |  the Free Software Foundation; either version 2, or (at your option)     |
 |  any later version.                                                      |
 |                                                                          |
 |  This program is distributed in the hope that it will be useful,         |
 |  but WITHOUT ANY WARRANTY; without even the implied warranty of          |
 |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           |
 |  GNU General Public License for more details.                            |
 |                                                                          |
 |  You should have received a copy of the GNU General Public License       |
 |  along with this program; if not, write to the Free Software             |
 |  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               |
 |                                                                          |
 |  Copyright (C) 2008                                                      |
 |                                                                          |
 |         , __                 , __                                        |
 |        /|/  \               /|/  \                                       |
 |         | __/ _   ,_         | __/ _   ,_                                | 
 |         |   \|/  /  |  |   | |   \|/  /  |  |   |                        |
 |         |(__/|__/   |_/ \_/|/|(__/|__/   |_/ \_/|/                       |
 |                           /|                   /|                        |
 |                           \|                   \|                        |
 |                                                                          |
 |      Enrico Bertolazzi                                                   |
 |      Dipartimento di Ingegneria Meccanica e Strutturale                  |
 |      Universita` degli Studi di Trento                                   |
 |      Via Mesiano 77, I-38050 Trento, Italy                               |
 |      email: enrico.bertolazzi@unitn.it                                   |
 |                                                                          |
\*--------------------------------------------------------------------------*/

///
/// file: alglin.hh
///

#ifndef ALGLIN_HH
#define ALGLIN_HH

#ifndef BLASNAME
  #define BLASNAME(A) A##_
#endif

#ifndef LAPACKNAME
  #define LAPACKNAME(A) A##_
#endif

#ifndef ATLASNAME
  #define ATLASNAME(A) ATL_##A
#endif

#include <stdexcept>
#include <cmath>

namespace alglin {

  using namespace std ;

  typedef char   character ;
  typedef int    integer ;
  typedef float  single_precision ;
  typedef double double_precision ;

  enum ATLAS_ORDER { AtlasRowMajor  = 101, AtlasColMajor = 102 };
  enum ATLAS_TRANS { AtlasNoTrans   = 111, AtlasTrans    = 112,
                     AtlasConjTrans = 113, AtlasConj     = 114 };
  enum ATLAS_UPLO  { AtlasUpper     = 121, AtlasLower    = 122 };
  enum ATLAS_DIAG  { AtlasNonUnit   = 131, AtlasUnit     = 132 };

  enum ATLAS_SIDE  { AtlasLeft=141, AtlasRight=142 } ;

  inline
  ATLAS_TRANS
  trans_atlas(character const TRANS ) {
    switch ( TRANS ) {
    case 'N': case 'n': return AtlasNoTrans ;
    case 'T': case 't': return AtlasTrans ;
    case 'C': case 'c': return AtlasConjTrans ;
    default:  throw runtime_error("bad traspose mode selection\n") ;
    }
  }

  inline
  ATLAS_UPLO
  uplo_atlas(character const UPLO ) {
    switch ( UPLO ) {
    case 'U': case 'u': return AtlasUpper ;
    case 'L': case 'l': return AtlasLower ;
    default:  throw runtime_error("bad UP/LOW mode selection\n") ;
    }
  }

  inline
  ATLAS_DIAG
  diag_atlas(character const DIAG ) {
    switch ( DIAG ) {
    case 'U': case 'u': return AtlasUnit ;
    case 'N': case 'n': return AtlasNonUnit ;
    default:  throw runtime_error("bad diagonal mode selection\n") ;
    }
  }
  
  inline
  ATLAS_SIDE
  side_atlas(character const SIDE ) {
    switch ( SIDE ) {
    case 'L': case 'l': return AtlasLeft ;
    case 'R': case 'r': return AtlasRight ;
    default:  throw runtime_error("bad side mode selection\n") ;
    }
  }

  /*
  //    ___ ___  _ __  _   _ 
  //   / __/ _ \| '_ \| | | |
  //  | (_| (_) | |_) | |_| |
  //   \___\___/| .__/ \__, |
  //            |_|    |___/ 
  */
  extern "C" {
    void
    BLASNAME(scopy)( integer          const * N, 
                     single_precision const   X[], 
                     integer          const * INCX, 
                     single_precision         Y[], 
                     integer          const * INCY ) ;
    void
    BLASNAME(dcopy)( integer          const * N, 
                     double_precision const   X[], 
                     integer          const * INCX, 
                     double_precision         Y[], 
                     integer          const * INCY ) ;
    void
    ATLASNAME(scopy)( const int N,
                      const float *X, const int incX,
                      float       *Y, const int incY );
    void
    ATLASNAME(dcopy)( const int N,
                      const double *X, const int incX,
                      double       *Y, const int incY );
  }

  inline 
  void
  copy( integer          const N, 
        single_precision const X[], 
        integer          const INCX, 
        single_precision       Y[], 
        integer          const INCY )
  #ifdef USE_ATLAS
  { ATLASNAME(scopy)( N, X, INCX, Y, INCY ) ; }
  #else
  { BLASNAME(scopy)( &N, X, &INCX, Y, &INCY ) ; }
  #endif

  inline
  void
  copy( integer          const N, 
        double_precision const X[], 
        integer          const INCX, 
        double_precision       Y[], 
        integer          const INCY )
  #ifdef USE_ATLAS
  { ATLASNAME(dcopy)( N, X, INCX, Y, INCY ) ; }
  #else
  { BLASNAME(dcopy)( &N, X, &INCX, Y, &INCY ) ; }
  #endif

  /*
  //   _____      ____ _ _ __  
  //  / __\ \ /\ / / _` | '_ \ 
  //  \__ \\ V  V / (_| | |_) |
  //  |___/ \_/\_/ \__,_| .__/ 
  //                    |_|    
  */
  extern "C" {
    void
    BLASNAME(sswap)( integer          const * N, 
                     single_precision         X[], 
                     integer          const * INCX, 
                     single_precision         Y[], 
                     integer          const * INCY ) ;
    void
    BLASNAME(dswap)( integer          const * N, 
                     double_precision         X[], 
                     integer          const * INCX, 
                     double_precision         Y[], 
                     integer          const * INCY ) ;
    void
    ATLASNAME(sswap)( const int N,
                      float *X, const int incX,
                      float *Y, const int incY );
    void
    ATLASNAME(dswap)( const int N,
                      double *X, const int incX,
                      double *Y, const int incY );
    void
    BLASNAME(slaswp)( integer          const * NCOL, 
                      single_precision         A[], 
                      integer          const * LDA, 
                      integer          const * K1, 
                      integer          const * K2,
                      integer          const   IPIV[],
                      integer          const * INC ) ;
    void
    BLASNAME(dlaswp)( integer          const * NCOL, 
                      double_precision         A[], 
                      integer          const * LDA, 
                      integer          const * K1, 
                      integer          const * K2,
                      integer          const   IPIV[],
                      integer          const * INC ) ;
  }

  inline
  void
  swap( integer          const N, 
        single_precision       X[], 
        integer          const INCX, 
        single_precision       Y[], 
        integer          const INCY )
  #ifdef USE_ATLAS
  { ATLASNAME(sswap)( N, X, INCX, Y, INCY ) ; }
  #else
  { BLASNAME(sswap)( &N, X, &INCX, Y, &INCY ) ; }
  #endif

  inline
  void
  swap( integer          const N, 
        double_precision       X[], 
        integer          const INCX, 
        double_precision       Y[], 
        integer          const INCY )
  #ifdef USE_ATLAS
  { ATLASNAME(dswap)( N, X, INCX, Y, INCY ) ; }
  #else
  { BLASNAME(dswap)( &N, X, &INCX, Y, &INCY ) ; }
  #endif

  inline
  void
  swaps( integer          const NCOL, 
         single_precision       A[], 
         integer          const LDA, 
         integer          const I1, 
         integer          const I2,
         integer          const IPIV[],
         integer          const INC ) {
    integer K1 = I1+1 ;
    integer K2 = I2+1 ;
    BLASNAME(slaswp)( &NCOL, A, &LDA, &K1, &K2, IPIV, &INC ) ;
  }

  inline
  void
  swaps( integer          const NCOL, 
         double_precision       A[], 
         integer          const LDA, 
         integer          const I1, 
         integer          const I2,
         integer          const IPIV[],
         integer          const INC ) {
    integer K1 = I1+1 ;
    integer K2 = I2+1 ;
    BLASNAME(dlaswp)( &NCOL, A, &LDA, &K1, &K2, IPIV, &INC ) ;
  }

  /*
  //                 _ 
  //   ___  ___ __ _| |
  //  / __|/ __/ _` | |
  //  \__ \ (_| (_| | |
  //  |___/\___\__,_|_|
  */
  extern "C" {
    void
    BLASNAME(sscal)( integer          const * N,
                     single_precision const * S,
                     single_precision         X[],
                     integer          const * INCX ) ;
    void
    BLASNAME(dscal)( integer          const * N,
                     double_precision const * S,
                     double_precision         X[],
                     integer          const * INCX ) ;
    void
    ATLASNAME(sscal)( const int N,
                      const float alpha,
                      float *X, const int incX );
    void
    ATLASNAME(dscal)( const int N,
                      const double alpha,
                      double *X, const int incX );
  }

  inline
  void
  scal( integer          const N,
        single_precision const S,
        single_precision       X[],
        integer          const INCX )
  #ifdef USE_ATLAS
  { ATLASNAME(sscal)(N, S, X, INCX) ; }
  #else
  { BLASNAME(sscal)(&N, &S, X, &INCX) ; }
  #endif

  inline
  void
  scal( integer          const N,
        double_precision const S,
        double_precision       X[],
        integer          const INCX )
  #ifdef USE_ATLAS
  { ATLASNAME(dscal)(N, S, X, INCX) ; }
  #else
  { BLASNAME(dscal)(&N, &S, X, &INCX) ; }
  #endif

  /*
  //    __ ___  ___ __  _   _ 
  //   / _` \ \/ / '_ \| | | |
  //  | (_| |>  <| |_) | |_| |
  //   \__,_/_/\_\ .__/ \__, |
  //             |_|    |___/
  */
  extern "C" {
    void
    BLASNAME(saxpy)( integer          const * N, 
                     single_precision const * A, 
                     single_precision const   X[], 
                     integer          const * INCX, 
                     single_precision         Y[], 
                     integer          const * INCY ) ;
    void
    BLASNAME(daxpy)( integer          const * N, 
                     double_precision const * A, 
                     double_precision const   X[], 
                     integer          const * INCX, 
                     double_precision         Y[], 
                     integer          const * INCY ) ;
    void
    ATLASNAME(saxpy)( const int N,
                      const float alpha,
                      const float *X, const int incX,
                      float       *Y, const int incY );
    void
    ATLASNAME(daxpy)( const int N,
                      const double alpha,
                      const double *X, const int incX,
                      double       *Y, const int incY );
  }

  inline
  void
  axpy( integer          const N, 
        single_precision const A, 
        single_precision const X[], 
        integer          const INCX, 
        single_precision       Y[], 
        integer          const INCY )
  #ifdef USE_ATLAS
  { ATLASNAME(saxpy)( N, A, X, INCX, Y, INCY ) ; }
  #else
  { BLASNAME(saxpy)( &N, &A, X, &INCX, Y, &INCY ) ; }
  #endif

  inline
  void
  axpy( integer          const N, 
        double_precision const A, 
        double_precision const X[], 
        integer          const INCX, 
        double_precision       Y[], 
        integer          const INCY )
  #ifdef USE_ATLAS
  { ATLASNAME(daxpy)( N, A, X, INCX, Y, INCY ) ; }
  #else
  { BLASNAME(daxpy)( &N, &A, X, &INCX, Y, &INCY ) ; }
  #endif

  /*
  //   _______ _ __ ___  
  //  |_  / _ \ '__/ _ \ 
  //   / /  __/ | | (_) |
  //  /___\___|_|  \___/ 
  */
  extern "C" {  
    void
    ATLASNAME(szero)( const int N, float *X, const int incX );
    void
    ATLASNAME(dzero)( const int N, double *X, const int incX );
  }

  inline
  void
  zero( integer          const N,
        single_precision       X[],
        integer          const INCX )
  #ifdef USE_ATLAS
  { ATLASNAME(szero)( N, X, INCX ) ; }
  #else
  { single_precision z  = 0 ;
    integer          iz = 0 ;
    BLASNAME(scopy)( &N, &z, &iz, X, &INCX ) ; }
  #endif

  inline
  void
  zero( integer          const N,
        double_precision       X[],
        integer          const INCX )
  #ifdef USE_ATLAS
  { ATLASNAME(dzero)( N, X, INCX ) ; }
  #else
  { double_precision z  = 0 ;
    integer          iz = 0 ;
    BLASNAME(dcopy)( &N, &z, &iz, X, &INCX ) ; }
  #endif

  /*
  //    __ _  ___ _ __ 
  //   / _` |/ _ \ '__|
  //  | (_| |  __/ |   
  //   \__, |\___|_|   
  //   |___/
  */
  extern "C" {
    void
    BLASNAME(sger)( integer          const * M, 
                    integer          const * N, 
                    single_precision const * ALPHA, 
                    single_precision const   X[], 
                    integer          const * INCX, 
                    single_precision const   Y[], 
                    integer          const * INCY, 
                    single_precision         A[], 
                    integer          const * LDA ) ;
    void
    BLASNAME(dger)( integer          const * M, 
                    integer          const * N, 
                    double_precision const * ALPHA, 
                    double_precision const   X[], 
                    integer          const * INCX, 
                    double_precision const   Y[], 
                    integer          const * INCY, 
                    double_precision         A[], 
                    integer          const * LDA ) ;
    void
    ATLASNAME(sger)( const int M, const int N,
                     const float alpha,
                     const float *X, const int incX,
                     const float *Y, const int incY,
                     float       *A, const int lda );
    void
    ATLASNAME(dger)( const int M, const int N,
                     const double alpha,
                     const double *X, const int incX,
                     const double *Y, const int incY,
                     double       *A, const int lda );
  }
  
  inline
  void
  ger( integer          const M, 
       integer          const N, 
       single_precision const ALPHA, 
       single_precision const X[], 
       integer          const INCX, 
       single_precision const Y[], 
       integer          const INCY, 
       single_precision       A[], 
       integer          const LDA )
  #ifdef USE_ATLAS
  { ATLASNAME(sger)( M, N, ALPHA, X, INCX, Y, INCY, A, LDA ) ; }
  #else
  { BLASNAME(sger)( &M, &N, &ALPHA, X, &INCX, Y, &INCY, A, &LDA ) ; }
  #endif

  inline
  void
  ger( integer          const M, 
       integer          const N, 
       double_precision const ALPHA, 
       double_precision const X[], 
       integer          const INCX, 
       double_precision const Y[], 
       integer          const INCY, 
       double_precision       A[], 
       integer          const LDA )
  #ifdef USE_ATLAS
  { ATLASNAME(dger)( M, N, ALPHA, X, INCX, Y, INCY, A, LDA ) ; }
  #else
  { BLASNAME(dger)( &M, &N, &ALPHA, X, &INCX, Y, &INCY, A, &LDA ) ; }
  #endif

  /*
  //    __ _  ___ _ __ _____   __
  //   / _` |/ _ \ '_ ` _ \ \ / /
  //  | (_| |  __/ | | | | \ V / 
  //   \__, |\___|_| |_| |_|\_/  
  //   |___/                     
  */
  extern "C" {
    void
    BLASNAME(sgemv)( character        const   TRANS[], 
                     integer          const * M, 
                     integer          const * N, 
                     single_precision const * ALPHA, 
                     single_precision const   A[], 
                     integer          const * LDA,
                     single_precision const   X[], 
                     integer          const * INCX, 
                     single_precision const * BETA, 
                     single_precision         Y[], 
                     integer          const * INCY ) ;
    void
    BLASNAME(dgemv)( character        const   TRANS[], 
                     integer          const * M, 
                     integer          const * N, 
                     double_precision const * ALPHA, 
                     double_precision const   A[], 
                     integer          const * LDA,
                     double_precision const   X[], 
                     integer          const * INCX, 
                     double_precision const * BETA, 
                     double_precision         Y[], 
                     integer          const * INCY ) ;
    void
    ATLASNAME(sgemv)( const enum ATLAS_TRANS TransA,
                      const int M, const int N,
                      const float alpha,
                      const float *A, const int lda,
                      const float *X, const int incX,
                      const float beta,
                      float *Y, const int incY );
    void
    ATLASNAME(dgemv)( const enum ATLAS_TRANS TransA,
                      const int M, const int N,
                      const double alpha,
                      const double *A, const int lda,
                      const double *X, const int incX,
                      const double beta,
                      double *Y, const int incY );
  }
  
  inline
  void
  gemv( character        const TRANS[], 
        integer          const M, 
        integer          const N, 
        single_precision const ALPHA, 
        single_precision const A[], 
        integer          const LDA,
        single_precision const X[], 
        integer          const INCX, 
        single_precision const BETA, 
        single_precision       Y[], 
        integer          const INCY )
  #ifdef USE_ATLAS
  { ATLASNAME(sgemv)( trans_atlas(TRANS[0]), M, N, ALPHA, A, LDA, X, INCX, BETA, Y, INCY ) ; }
  #else
  { BLASNAME(sgemv)( TRANS, &M, &N, &ALPHA, A, &LDA, X, &INCX, &BETA, Y, &INCY ) ; }
  #endif

  inline
  void
  gemv( character        const TRANS[], 
        integer          const M, 
        integer          const N, 
        double_precision const ALPHA, 
        double_precision const A[], 
        integer          const LDA,
        double_precision const X[], 
        integer          const INCX, 
        double_precision const BETA, 
        double_precision       Y[], 
        integer          const INCY )
  #ifdef USE_ATLAS
  { ATLASNAME(dgemv)( trans_atlas(TRANS[0]), M, N, ALPHA, A, LDA, X, INCX, BETA, Y, INCY ) ; }
  #else
  { BLASNAME(dgemv)( TRANS, &M, &N, &ALPHA, A, &LDA, X, &INCX, &BETA, Y, &INCY ) ; }
  #endif

  /*
  //    __ _  ___ _ __ ___  _ __ ___  
  //   / _` |/ _ \ '_ ` _ \| '_ ` _ \ 
  //  | (_| |  __/ | | | | | | | | | |
  //   \__, |\___|_| |_| |_|_| |_| |_|
  //   |___/
  */
  extern "C" {
    void
    BLASNAME(sgemm)( character        const   TRANSA[], 
                     character        const   TRANSB[], 
                     integer          const * M, 
                     integer          const * N, 
                     integer          const * K, 
                     single_precision const * ALPHA, 
                     single_precision const   A[], 
                     integer          const * LDA,
                     single_precision const   B[], 
                     integer          const * LDB,
                     single_precision const * BETA, 
                     single_precision         C[], 
                     integer          const * LDC ) ;
    void
    BLASNAME(dgemm)( character        const   TRANSA[], 
                     character        const   TRANSB[], 
                     integer          const * M, 
                     integer          const * N, 
                     integer          const * K, 
                     double_precision const * ALPHA, 
                     double_precision const   A[], 
                     integer          const * LDA,
                     double_precision const   B[], 
                     integer          const * LDB,
                     double_precision const * BETA, 
                     double_precision         C[], 
                     integer          const * LDC ) ;
    void
    ATLASNAME(sgemm)( const enum ATLAS_TRANS TransA,
                      const enum ATLAS_TRANS TransB,
                      const int M, const int N, const int K,
                      const float alpha,
                      const float *A, const int lda,
                      const float *B, const int ldb,
                      const float beta,
                            float *C, const int ldc );
    void
    ATLASNAME(dgemm)( const enum ATLAS_TRANS TransA,
                      const enum ATLAS_TRANS TransB,
                      const int M, const int N, const int K,
                      const double alpha,
                      const double *A, const int lda,
                      const double *B, const int ldb,
                      const double beta,
                            double *C, const int ldc );
  }
  
  inline
  void
  gemm( character        const TRANSA[], 
        character        const TRANSB[], 
        integer          const M, 
        integer          const N, 
        integer          const K, 
        single_precision const ALPHA, 
        single_precision const A[], 
        integer          const LDA,
        single_precision const B[], 
        integer          const LDB,
        single_precision const BETA, 
        single_precision       C[], 
        integer          const LDC )
  #ifdef USE_ATLAS
  { ATLASNAME(sgemm)( trans_atlas(TRANSA[0]), trans_atlas(TRANSB[0]),
                      M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC ) ; }
  #else
  { BLASNAME(sgemm)( TRANSA, TRANSB,
                    &M, &N, &K,
                    &ALPHA, A, &LDA,
                    B, &LDB,
                    &BETA, C, &LDC ) ; }
  #endif

  inline
  void
  gemm( character        const TRANSA[],
        character        const TRANSB[], 
        integer          const M, 
        integer          const N, 
        integer          const K, 
        double_precision const ALPHA,
        double_precision const A[], 
        integer          const LDA,
        double_precision const B[], 
        integer          const LDB,
        double_precision const BETA, 
        double_precision       C[], 
        integer          const LDC )
  #ifdef USE_ATLAS
  { ATLASNAME(dgemm)( trans_atlas(TRANSA[0]), trans_atlas(TRANSB[0]),
                      M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC ) ; }
  #else
  { BLASNAME(dgemm)( TRANSA, TRANSB,
                    &M, &N, &K,
                    &ALPHA, A, &LDA,
                    B, &LDB,
                    &BETA, C, &LDC ) ; }
  #endif

  /*
  //                        ____  
  //   _ __  _ __ _ __ ___ |___ \ 
  //  | '_ \| '__| '_ ` _ \  __) |
  //  | | | | |  | | | | | |/ __/ 
  //  |_| |_|_|  |_| |_| |_|_____|
  */
  extern "C" {
    single_precision
    BLASNAME(snrm2)( integer const * N, single_precision const X[], integer const * INCX ) ;

    double_precision
    BLASNAME(dnrm2)( integer const * N, double_precision const X[], integer const * INCX ) ;

    float
    ATLASNAME(snrm2)( const int N, const float *X, const int incX ) ;

    double
    ATLASNAME(dnrm2)( const int N, const double *X, const int incX ) ;
  }

  inline
  single_precision
  nrm2( integer          const N,
        single_precision const X[],
        integer          const INCX )
  #ifdef USE_ATLAS
  { return ATLASNAME(snrm2)( N, X, INCX ) ; }
  #else
  { return BLASNAME(snrm2)( &N, X, &INCX ) ; }
  #endif

  inline
  double_precision
  nrm2( integer          const N,
        double_precision const X[],
        integer          const INCX )
  #ifdef USE_ATLAS
  { return ATLASNAME(dnrm2)( N, X, INCX ) ; }
  #else
  { return BLASNAME(dnrm2)( &N, X, &INCX ) ; }
  #endif

  /*
  //    __ _ ___ _   _ _ __ ___  
  //   / _` / __| | | | '_ ` _ \ 
  //  | (_| \__ \ |_| | | | | | |
  //   \__,_|___/\__,_|_| |_| |_|
  */
  extern "C" {

    single_precision
    BLASNAME(sasum)( integer const * N, single_precision const X[], integer const * INCX ) ;

    double_precision
    BLASNAME(dasum)( integer const * N, double_precision const X[], integer const * INCX ) ;

    float
    ATLASNAME(sasum)( const int N, const float *X, const int incX );

    double
    ATLASNAME(dasum)( const int N, const double *X, const int incX );
  }                         

  inline
  single_precision
  asum(integer          const N,
       single_precision const X[],
       integer          const INCX)
  #ifdef USE_ATLAS
  { return ATLASNAME(sasum)( N, X, INCX ) ; }
  #else
  { return BLASNAME(sasum)( &N, X, &INCX ) ; }
  #endif
  
  inline
  double_precision
  asum(integer          const N,
       double_precision const X[],
       integer          const INCX)
  #ifdef USE_ATLAS
  { return ATLASNAME(dasum)( N, X, INCX ) ; }
  #else
  { return BLASNAME(dasum)( &N, X, &INCX ) ; }
  #endif

  /*
  //    __ _ _ __ ___   __ ___  __
  //   / _` | '_ ` _ \ / _` \ \/ /
  //  | (_| | | | | | | (_| |>  < 
  //   \__,_|_| |_| |_|\__,_/_/\_\
  */
  extern "C" {

    integer
    BLASNAME(isamax)( integer const * N, single_precision const X[], integer const * INCX ) ;

    integer
    BLASNAME(idamax)( integer const * N, double_precision const X[], integer const * INCX ) ;

    int
    ATLASNAME(isamax)( const int N, const float *X, const int incX );

    int
    ATLASNAME(idamax)( const int N, const double *X, const int incX );  
  }
  
  inline
  integer
  iamax( integer          const N,
         single_precision const X[],
         integer          const INCX )
  #ifdef USE_ATLAS
  { return ATLASNAME(isamax)( N, X, INCX )-1 ; }
  #else
  { return BLASNAME(isamax)( &N, X, &INCX )-1 ; }
  #endif

  inline
  integer
  iamax( integer          const N,
         double_precision const X[],
         integer          const INCX )
  #ifdef USE_ATLAS
  { return ATLASNAME(idamax)( N, X, INCX )-1 ; }
  #else
  { return BLASNAME(idamax)( &N, X, &INCX )-1 ; }
  #endif

  inline
  single_precision
  amax( integer          const N,
        single_precision const X[],
        integer          const INCX )
  { return abs(X[iamax(N,X,INCX)]) ; }
  
  inline
  double_precision
  amax( integer          const N,
        double_precision const X[],
        integer          const INCX )
  { return abs(X[iamax(N,X,INCX)]) ; }

  /*
  //       _       _   
  //    __| | ___ | |_ 
  //   / _` |/ _ \| __|
  //  | (_| | (_) | |_ 
  //   \__,_|\___/ \__|
  */
  extern "C" {
    single_precision
    BLASNAME(sdot)( integer          const * N, 
                    single_precision const   SX[], 
                    integer          const * INCX, 
                    single_precision const   SY[], 
                    integer          const * INCY ) ;

    double_precision
    BLASNAME(ddot)( integer          const * N, 
                    double_precision const   SX[], 
                    integer          const * INCX, 
                    double_precision const   SY[], 
                    integer          const * INCY ) ;
    float
    ATLASNAME(sdot)( const int N,
                     const float *X, const int incX,
                     const float *Y, const int incY );
    double
    ATLASNAME(ddot)( const int N,
                     const double *X, const int incX,
                     const double *Y, const int incY );
  }

  inline
  single_precision
  dot( integer          const N, 
       single_precision const SX[], 
       integer          const INCX, 
       single_precision const SY[], 
       integer          const INCY )
  #ifdef USE_ATLAS
  { return ATLASNAME(sdot)( N, SX, INCX, SY, INCY ) ; }
  #else
  { return BLASNAME(sdot)( &N, SX, &INCX, SY, &INCY ) ; }
  #endif
  
  inline
  double_precision
  dot( integer          const N, 
       double_precision const SX[], 
       integer          const INCX, 
       double_precision const SY[], 
       integer          const INCY ) 
  #ifdef USE_ATLAS
  { return ATLASNAME(ddot)( N, SX, INCX, SY, INCY ) ; }
  #else
  { return BLASNAME(ddot)( &N, SX, &INCX, SY, &INCY ) ; }
  #endif

  /*
  //    __ _  ___  ___ ___  _ __  _   _ 
  //   / _` |/ _ \/ __/ _ \| '_ \| | | |
  //  | (_| |  __/ (_| (_) | |_) | |_| |
  //   \__, |\___|\___\___/| .__/ \__, |
  //   |___/               |_|    |___/ 
  */
  extern "C" {
    void
    LAPACKNAME(slacpy)( character        const   UPLO[], 
                        integer          const * M, 
                        integer          const * N, 
                        single_precision const   A[], 
                        integer          const * LDA, 
                        single_precision         B[], 
                        integer          const * LDB ) ;
    void
    LAPACKNAME(dlacpy)( character        const   UPLO[], 
                        integer          const * M, 
                        integer          const * N, 
                        double_precision const   A[], 
                        integer          const * LDA, 
                        double_precision         B[], 
                        integer          const * LDB ) ;
    void
    ATLASNAME(sgecopy)( const int M, const int N,
                        const float *A, const int lda,
                        float       *C, const int ldc ) ;
    void
    ATLASNAME(dgecopy)( const int M, const int N,
                        const double *A, const int lda,
                        double       *C, const int ldc ) ;
  }

  inline
  void
  gecopy( integer          const M, 
          integer          const N, 
          single_precision const A[], 
          integer          const LDA, 
          single_precision       B[], 
          integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(sgecopy)( M, N, A, LDA, B, LDB ) ; }
  #else
  { LAPACKNAME(slacpy)( "A", &M, &N, A, &LDA, B, &LDB ) ; }
  #endif

  inline
  void
  gecopy( integer          const M, 
          integer          const N, 
          double_precision const A[], 
          integer          const LDA, 
          double_precision       B[], 
          integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(dgecopy)( M, N, A, LDA, B, LDB ) ; }
  #else
  { LAPACKNAME(dlacpy)( "A", &M, &N, A, &LDA, B, &LDB ) ; }
  #endif

  /*
  //    __ _  ___ _______ _ __ ___  
  //   / _` |/ _ \_  / _ \ '__/ _ \ 
  //  | (_| |  __// /  __/ | | (_) |
  //   \__, |\___/___\___|_|  \___/ 
  //   |___/                        
  */
  extern "C" {
    void
    LAPACKNAME(slaset)( character        const   UPLO[], 
                        integer          const * M, 
                        integer          const * N, 
                        single_precision const * ALPHA, 
                        single_precision const * BETA, 
                        single_precision         A[], 
                        integer          const * LDA ) ;
    void
    LAPACKNAME(dlaset)( character        const   UPLO[], 
                        integer          const * M, 
                        integer          const * N, 
                        double_precision const * ALPHA, 
                        double_precision const * BETA, 
                        double_precision         A[], 
                        integer          const * LDA ) ;
    void
    ATLASNAME(sgezero)( const int M, const int N, float *C, const int ldc ) ;    
    void
    ATLASNAME(dgezero)( const int M, const int N, double *C, const int ldc ) ;

  }

  inline
  void
  gezero( integer          const M,
          integer          const N, 
          single_precision       A[], 
          integer          const LDA )
  #ifdef USE_ATLAS
  { ATLASNAME(sgezero)( M, N, A, LDA ) ; }
  #else
  { single_precision zero = 0 ;
    LAPACKNAME(slaset)( "A", &M, &N, &zero, &zero, A, &LDA ) ; }
  #endif

  inline
  void
  gezero( integer          const M, 
          integer          const N, 
          double_precision       A[], 
          integer          const LDA )
  #ifdef USE_ATLAS
  { ATLASNAME(dgezero)( M, N, A, LDA ) ; }
  #else
  { double_precision zero = 0 ;
    LAPACKNAME(dlaset)( "A", &M, &N, &zero, &zero, A, &LDA ) ; }
  #endif
  
  /*
  //                        _     _ 
  //    __ _  ___  __ _  __| | __| |
  //   / _` |/ _ \/ _` |/ _` |/ _` |
  //  | (_| |  __/ (_| | (_| | (_| |
  //   \__, |\___|\__,_|\__,_|\__,_|
  //   |___/                        
  */
  extern "C" {
    void
    ATLASNAME(sgeadd)( const int M, const int N,
                       const float alpha, const float *A, const int lda,
                       const float beta,        float *C, const int ldc ) ;
    void
    ATLASNAME(dgeadd)( const int M, const int N,
                       const double alpha, const double *A, const int lda,
                       const double beta,        double *C, const int ldc ) ;
  }
  /*
   * B <- alpha*A + beta*B
   */
  inline
  void
  geadd( integer          const M,
         integer          const N, 
         single_precision const alpha,
         single_precision const A[], 
         integer          const LDA,
         single_precision const beta,
         single_precision       B[], 
         integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(sgeadd)( M, N, alpha, A, LDA, beta, B, LDB ) ; }
  #else
  { for ( integer j = 0 ; j < N ; ++j ) {
     single_precision const * Aj = A + j * LDA ;
     single_precision       * Bj = B + j * LDB ;
     alglin::scal( M, beta,  Bj, 1 ) ;
     alglin::axpy( M, alpha, Aj, 1, Bj, 1 ) ;
    }
  }
  #endif

  inline
  void
  geadd( integer          const M,
         integer          const N, 
         double_precision const alpha,
         double_precision const A[], 
         integer          const LDA,
         double_precision const beta,
         double_precision       B[], 
         integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(dgeadd)( M, N, alpha, A, LDA, beta, B, LDB ) ; }
  #else
  { for ( integer j = 0 ; j < N ; ++j ) {
     double_precision const * Aj = A + j * LDA ;
     double_precision       * Bj = B + j * LDB ;
     alglin::scal( M, beta,  Bj, 1 ) ;
     alglin::axpy( M, alpha, Aj, 1, Bj, 1 ) ;
    }
  }
  #endif

  /*
  //              _         __ 
  //    __ _  ___| |_ _ __ / _|
  //   / _` |/ _ \ __| '__| |_ 
  //  | (_| |  __/ |_| |  |  _|
  //   \__, |\___|\__|_|  |_|  
  //   |___/                   
  */
  extern "C" {  
    void
    LAPACKNAME(sgetrf)( integer          const * N, 
                        integer          const * M, 
                        single_precision         A[], 
                        integer          const * LDA, 
                        integer                  IPIV[], 
                        integer                * INFO ) ;
    void
    LAPACKNAME(dgetrf)( integer          const * N, 
                        integer          const * M, 
                        double_precision         A[], 
                        integer          const * LDA, 
                        integer                  IPIV[], 
                        integer                * INFO ) ;
    int
    ATLASNAME(sgetrf)( const enum ATLAS_ORDER Order,
                       const int M, const int N,
                       float *A, const int lda,
                       int *ipiv ) ;
    int
    ATLASNAME(dgetrf)( const enum ATLAS_ORDER Order,
                       const int M, const int N,
                       double *A, const int lda,
                       int *ipiv ) ;
  }
  
  inline
  integer
  getrf( integer          const N, 
         integer          const M, 
         single_precision       A[], 
         integer          const LDA, 
         integer                IPIV[] )
  #ifdef USE_ATLAS
  { return ATLASNAME(sgetrf)( AtlasColMajor, N, M, A, LDA, IPIV ) ; }
  #else
  { integer INFO ;
    LAPACKNAME(sgetrf)( &N, &M, A, &LDA, IPIV, &INFO ) ;
    return INFO ; }
  #endif

  inline
  integer
  getrf( integer          const N,
         integer          const M,
         double_precision       A[],
         integer          const LDA, 
         integer                IPIV[] )
  #ifdef USE_ATLAS
  { return ATLASNAME(dgetrf)( AtlasColMajor, N, M, A, LDA, IPIV ) ; }
  #else
  { integer INFO ;
    LAPACKNAME(dgetrf)( &N, &M, A, &LDA, IPIV, &INFO ) ;
    return INFO ; }
  #endif

  /*
  //              _            
  //    __ _  ___| |_ _ __ ___ 
  //   / _` |/ _ \ __| '__/ __|
  //  | (_| |  __/ |_| |  \__ \
  //   \__, |\___|\__|_|  |___/
  //   |___/                   
  */
  extern "C" {
    void
    LAPACKNAME(sgetrs)( character        const   TRANS[], 
                        integer          const * N, 
                        integer          const * NRHS, 
                        single_precision const   A[], 
                        integer          const * LDA, 
                        integer          const   IPIV[], 
                        single_precision         B[], 
                        integer          const * LDB, 
                        integer                * INFO ) ;
    void
    LAPACKNAME(dgetrs)( character        const   TRANS[],
                        integer          const * N, 
                        integer          const * NRHS, 
                        double_precision const   A[], 
                        integer          const * LDA, 
                        integer          const   IPIV[], 
                        double_precision         B[], 
                        integer          const * LDB, 
                        integer                * INFO ) ;
    void
    ATLASNAME(sgetrs)( const enum ATLAS_ORDER Order,
                       const enum ATLAS_TRANS Trans,
                       const int N,
                       const int NRHS,
                       const float *A, const int lda,
                       const int *ipiv,
                       float *B, const int ldb );
    void
    ATLASNAME(dgetrs)( const enum ATLAS_ORDER Order,
                       const enum ATLAS_TRANS Trans,
                       const int N,
                       const int NRHS,
                       const double *A, const int lda,
                       const int *ipiv,
                       double *B, const int ldb );
  }
  
  inline
  integer
  getrs( character        const TRANS[], 
         integer          const N, 
         integer          const NRHS, 
         single_precision const A[], 
         integer          const LDA, 
         integer          const IPIV[], 
         single_precision       B[], 
         integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(sgetrs)( AtlasColMajor, trans_atlas(TRANS[0]), N, NRHS, A, LDA, IPIV, B, LDB ) ;
    return 0 ; }
  #else
  { integer INFO ; 
    LAPACKNAME(sgetrs)( TRANS, &N, &NRHS, A, &LDA, IPIV, B, &LDB, &INFO ) ;
    return INFO ; }
  #endif

  inline
  integer
  getrs( character        const TRANS[], 
         integer          const N, 
         integer          const NRHS, 
         double_precision const A[], 
         integer          const LDA, 
         integer          const IPIV[], 
         double_precision       B[], 
         integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(dgetrs)( AtlasColMajor, trans_atlas(TRANS[0]), N, NRHS, A, LDA, IPIV, B, LDB ) ;
    return 0 ; }
  #else
  { integer INFO ;
    LAPACKNAME(dgetrs)( TRANS, &N, &NRHS, A, &LDA, IPIV, B, &LDB, &INFO ) ;
    return INFO ; }
  #endif

  /*
  //    __ _  ___  _____   __
  //   / _` |/ _ \/ __\ \ / /
  //  | (_| |  __/\__ \\ V / 
  //   \__, |\___||___/ \_/  
  //   |___/                 
  */
  extern "C" {
    void 
    LAPACKNAME(sgesv)( integer          const * N, 
                       integer          const * NRHS, 
                       single_precision         A[], 
                       integer          const * LDA, 
                       integer                  IPIV[], 
                       single_precision         B[], 
                       integer          const * LDB, 
                       integer                * INFO ) ;
    void
    LAPACKNAME(dgesv)( integer          const * N, 
                       integer          const * NRHS, 
                       double_precision         A[], 
                       integer          const * LDA, 
                       integer                  IPIV[], 
                       double_precision         B[], 
                       integer          const * LDB, 
                       integer                * INFO ) ;
  }
  
  inline
  integer
  gesv( integer          const N, 
        integer          const NRHS, 
        single_precision       A[], 
        integer          const LDA, 
        integer                IPIV[], 
        single_precision       B[], 
        integer          const LDB )
  #ifdef USE_ATLAS
  { integer INFO = ATLASNAME(sgetrf)(AtlasColMajor, N, N, A, LDA, IPIV);
    if ( INFO == 0 ) ATLASNAME(sgetrs)(AtlasColMajor, AtlasNoTrans,
                                       N, NRHS, A, LDA, IPIV, B, LDB) ;
    return INFO ; }
  #else
  { integer INFO ;
    LAPACKNAME(sgesv)( &N, &NRHS, A, &LDA, IPIV, B, &LDB, &INFO ) ;
    return INFO ; }
  #endif

  inline
  integer
  gesv( integer          const N, 
        integer          const NRHS, 
        double_precision       A[], 
        integer          const LDA, 
        integer                IPIV[], 
        double_precision       B[], 
        integer          const LDB )
  #ifdef USE_ATLAS
  { integer INFO = ATLASNAME(dgetrf)(AtlasColMajor, N, N, A, LDA, IPIV);
    if ( INFO == 0 ) ATLASNAME(dgetrs)(AtlasColMajor, AtlasNoTrans,
                                       N, NRHS, A, LDA, IPIV, B, LDB) ;
    return INFO ; }
  #else
  { integer INFO ;
    LAPACKNAME(dgesv)( &N, &NRHS, A, &LDA, IPIV, B, &LDB, &INFO ) ;
    return INFO ; }
  #endif

  /*
  //         _   _         __ 
  //    __ _| |_| |_ _ __ / _|
  //   / _` | __| __| '__| |_ 
  //  | (_| | |_| |_| |  |  _|
  //   \__, |\__|\__|_|  |_|  
  //   |___/                  
  */
  extern "C" {
    void
    LAPACKNAME(sgttrf)( integer          const * N,
                        single_precision         DL[], 
                        single_precision         D[], 
                        single_precision         DU[], 
                        single_precision         DU2[], 
                        integer                  IPIV[], 
                        integer                * INFO ) ;
    void
    LAPACKNAME(dgttrf)( integer          const * N,
                        double_precision         DL[], 
                        double_precision         D[], 
                        double_precision         DU[], 
                        double_precision         DU2[], 
                        integer                  IPIV[], 
                        integer                * INFO ) ;
  }

  inline
  integer
  gttrf( integer  const   N,
         single_precision DL[], 
         single_precision D[], 
         single_precision DU[], 
         single_precision DU2[], 
         integer          IPIV[] )
  { integer INFO ;
    LAPACKNAME(sgttrf)( &N, DL, D, DU, DU2, IPIV, &INFO) ;
    return INFO ;  }

  inline
  integer
  gttrf( integer  const   N,
         double_precision DL[], 
         double_precision D[], 
         double_precision DU[], 
         double_precision DU2[], 
         integer          IPIV[] )
  { integer INFO ;
    LAPACKNAME(dgttrf)( &N, DL, D, DU, DU2, IPIV, &INFO) ;
    return INFO ; }
    
  /*
  //         _   _            
  //    __ _| |_| |_ _ __ ___ 
  //   / _` | __| __| '__/ __|
  //  | (_| | |_| |_| |  \__ \
  //   \__, |\__|\__|_|  |___/
  //   |___/                  
  */
  extern "C" {
    void
    LAPACKNAME(sgttrs)( character                TRANS[],
                        integer          const * N,
                        integer          const * NRHS, 
                        single_precision const   DL[], 
                        single_precision const   D[], 
                        single_precision const   DU[], 
                        single_precision const   DU2[], 
                        integer          const   IPIV[], 
                        single_precision         B[], 
                        integer          const * LDB, 
                        integer                * INFO ) ;

    void
    LAPACKNAME(dgttrs)( character                TRANS[],
                        integer          const * N,
                        integer          const * NRHS, 
                        double_precision const   DL[], 
                        double_precision const   D[], 
                        double_precision const   DU[], 
                        double_precision const   DU2[], 
                        integer          const   IPIV[], 
                        double_precision         B[], 
                        integer          const * LDB, 
                        integer                * INFO ) ;

  }

  inline
  integer
  gttrs( character              TRANS[],
         integer          const N,
         integer          const NRHS, 
         single_precision const DL[], 
         single_precision const D[], 
         single_precision const DU[], 
         single_precision const DU2[], 
         integer          const IPIV[], 
         single_precision       B[], 
         integer          const LDB )
  { integer INFO ;
    LAPACKNAME(sgttrs)( TRANS, &N, &NRHS, DL, D, DU, DU2, IPIV, B, &LDB, &INFO) ;
    return INFO ; }

  inline
  integer
  gttrs( character              TRANS[],
         integer          const N,
         integer          const NRHS, 
         double_precision const DL[], 
         double_precision const D[], 
         double_precision const DU[], 
         double_precision const DU2[], 
         integer          const IPIV[], 
         double_precision       B[], 
         integer          const LDB )
  { integer INFO ;
    LAPACKNAME(dgttrs)( TRANS, &N, &NRHS, DL, D, DU, DU2, IPIV, B, &LDB, &INFO) ;
    return INFO ; }

  /*
  //         _             
  //    __ _| |_ _____   __
  //   / _` | __/ __\ \ / /
  //  | (_| | |_\__ \\ V / 
  //   \__, |\__|___/ \_/  
  //   |___/               
  */
  extern "C" {
    void
    LAPACKNAME(sgtsv)( integer          const * N, 
                       integer          const * NRHS, 
                       single_precision         DL[], 
                       single_precision         D[], 
                       single_precision         DU[], 
                       single_precision         B[],
                       integer          const * LDB, 
                       integer                * INFO ) ;

    void 
    LAPACKNAME(dgtsv)( integer          const * N, 
                       integer          const * NRHS, 
                       double_precision         DL[], 
                       double_precision         D[], 
                       double_precision         DU[], 
                       double_precision         B[], 
                       integer          const * LDB, 
                       integer                * INFO ) ;
  }

  inline
  integer
  gtsv( integer  const   N,
        integer  const   NRHS,
        single_precision DL[],
        single_precision D[],
        single_precision DU[],
        single_precision B[], 
        integer  const   LDB )
  { integer INFO ;
    LAPACKNAME(sgtsv)( &N, &NRHS, DL, D, DU, B, &LDB, &INFO ) ;
    return INFO ; }

  inline
  integer
  gtsv( integer  const   N,
        integer  const   NRHS,
        double_precision DL[],
        double_precision D[],
        double_precision DU[],
        double_precision B[], 
        integer  const   LDB )
  { integer INFO ; 
    LAPACKNAME(dgtsv)( &N, &NRHS, DL, D, DU, B, &LDB, &INFO ) ;
    return INFO ; }

  /*
  //                              _ 
  //   _ __   ___  _ __ _ __ ___ / |
  //  | '_ \ / _ \| '__| '_ ` _ \| |
  //  | | | | (_) | |  | | | | | | |
  //  |_| |_|\___/|_|  |_| |_| |_|_|
  */
  extern "C" {

    // 'M' or 'm'            maxabs
    // '1', 'O' or 'o'       norm1
    // 'I' or 'i'            normI 
    // 'F', 'f', 'E' or 'e'  normF

    double_precision
    LAPACKNAME(dlange)( char             const   NORM[],
                        integer          const * M,
                        integer          const * N,
                        double_precision const * A,
                        integer          const * LDA,
                        double_precision       * WORK ) ;

    single_precision
    LAPACKNAME(slange)( char             const   NORM[],
                        integer          const * M,
                        integer          const * N,
                        single_precision const * A,
                        integer          const * LDA,
                        single_precision       * WORK ) ;
  
  }

  inline
  single_precision
  normI( integer          const N,
         integer          const M,
         single_precision const A[],
         integer          const LDA,
         single_precision       WORK[] )
  { return LAPACKNAME(slange)( "I", &N, &M, A, &LDA, WORK ) ; }

  inline
  double_precision
  normI( integer          const N,
         integer          const M,
         double_precision const A[],
         integer          const LDA,
         double_precision       WORK[] )
  { return LAPACKNAME(dlange)( "I", &N, &M, A, &LDA, WORK ) ; }

  ///////////////////

  inline
  single_precision
  norm1( integer          const N,
         integer          const M,
         single_precision const A[],
         integer          const LDA )
  { return LAPACKNAME(slange)( "1", &N, &M, A, &LDA, NULL ) ; }

  inline
  double_precision
  norm1( integer          const N,
         integer          const M,
         double_precision const A[],
         integer          const LDA )
  { return LAPACKNAME(dlange)( "1", &N, &M, A, &LDA, NULL ) ; }

  ///////////////////

  inline
  single_precision
  normF( integer          const N,
         integer          const M,
         single_precision const A[],
         integer          const LDA )
  { return LAPACKNAME(slange)( "F", &N, &M, A, &LDA, NULL ) ; }

  inline
  double_precision
  normF( integer          const N,
         integer          const M,
         double_precision const A[],
         integer          const LDA )
  { return LAPACKNAME(dlange)( "F", &N, &M, A, &LDA, NULL ) ; }

  ///////////////////

  inline
  single_precision
  maxabs( integer          const N,
          integer          const M,
          single_precision const A[],
          integer          const LDA )
  { return LAPACKNAME(slange)( "M", &N, &M, A, &LDA, NULL ) ; }

  inline
  double_precision
  maxabs( integer          const N,
          integer          const M,
          double_precision const A[],
          integer          const LDA )
  { return LAPACKNAME(dlange)( "M", &N, &M, A, &LDA, NULL ) ; }
  
  /*
  //   _                       
  //  | |_ _ __ _ __ _____   __
  //  | __| '__| '_ ` _ \ \ / /
  //  | |_| |  | | | | | \ V / 
  //   \__|_|  |_| |_| |_|\_/  
  */
  extern "C" {
    void
    BLASNAME(strmv)( char             const   UPLO[],
                     char             const   TRANS[],
                     char             const   DIAG[],
                     integer          const * N,
                     single_precision const   A[],
                     integer          const * LDA,
                     single_precision         X[],
                     integer          const * INCX ) ;
    void
    BLASNAME(dtrmv)( char             const   UPLO[],
                     char             const   TRANS[],
                     char             const   DIAG[],
                     integer          const * N,
                     double_precision const   A[],
                     integer          const * LDA,
                     double_precision         X[],
                     integer          const * INCX ) ;
    void
    ATLASNAME(strmv)( const enum ATLAS_UPLO  Uplo,
                      const enum ATLAS_TRANS TransA,
                      const enum ATLAS_DIAG  Diag,
                      const int N,
                      const float *A, const int lda,
                            float *X, const int incX ) ;
    void
    ATLASNAME(dtrmv)( const enum ATLAS_UPLO  Uplo,
                      const enum ATLAS_TRANS TransA,
                      const enum ATLAS_DIAG  Diag,
                      const int N,
                      const double *A, const int lda,
                            double *X, const int incX ) ;
  }

  inline
  void
  trmv( char             const UPLO[],
        char             const TRANS[],
        char             const DIAG[],
        integer          const N,
        single_precision const A[],
        integer          const LDA,
        single_precision       X[],
        integer          const INCX )
  #ifdef USE_ATLAS
  { ATLASNAME(strmv)( uplo_atlas(UPLO[0]),
                      trans_atlas(TRANS[0]),
                      diag_atlas(DIAG[0]),
                      N, A, LDA, X, INCX ) ; }
  #else
  { BLASNAME(strmv)( UPLO, TRANS, DIAG, &N, A, &LDA, X, &INCX ) ; }
  #endif

  inline
  void
  trmv( char             const UPLO[],
        char             const TRANS[],
        char             const DIAG[],
        integer          const N,
        double_precision const A[],
        integer          const LDA,
        double_precision       X[],
        integer          const INCX )
  #ifdef USE_ATLAS
  { ATLASNAME(dtrmv)( uplo_atlas(UPLO[0]),
                      trans_atlas(TRANS[0]),
                      diag_atlas(DIAG[0]),
                      N, A, LDA, X, INCX ) ; }
  #else
  { BLASNAME(dtrmv)( UPLO, TRANS, DIAG, &N, A, &LDA, X, &INCX ) ; }
  #endif

  /*
  //   _                  
  //  | |_ _ __ _____   __
  //  | __| '__/ __\ \ / /
  //  | |_| |  \__ \\ V / 
  //   \__|_|  |___/ \_/  
  */
  extern "C" {
    void
    BLASNAME(strsv)( character        const   UPLO[], 
                     character        const   TRANS[], 
                     character        const   DIAG[], 
                     integer          const * N, 
                     single_precision const   A[], 
                     integer          const * LDA, 
                     single_precision         X[], 
                     integer          const * INCX ) ;
    void
    BLASNAME(dtrsv)( character        const   UPLO[], 
                     character        const   TRANS[], 
                     character        const   DIAG[], 
                     integer          const * N, 
                     double_precision const   A[], 
                     integer          const * LDA, 
                     double_precision         X[], 
                     integer          const * INCX ) ;
    void
    ATLASNAME(strsv)( const enum ATLAS_UPLO  Uplo,
                      const enum ATLAS_TRANS TransA,
                      const enum ATLAS_DIAG  Diag,
                      const int N,
                      const float *A, const int lda,
                            float *X, const int incX ) ;
    void
    ATLASNAME(dtrsv)( const enum ATLAS_UPLO  Uplo,
                      const enum ATLAS_TRANS TransA,
                      const enum ATLAS_DIAG  Diag,
                      const int N,
                      const double *A, const int lda,
                            double *X, const int incX );
  }

  inline
  void
  trsv( character        const UPLO[], 
        character        const TRANS[], 
        character        const DIAG[],
        integer          const N, 
        single_precision const A[], 
        integer          const LDA, 
        single_precision       X[], 
        integer          const INCX )
  #ifdef USE_ATLAS
  { ATLASNAME(strsv)( uplo_atlas(UPLO[0]),
                      trans_atlas(TRANS[0]),
                      diag_atlas(DIAG[0]),
                      N, A, LDA, X, INCX ) ; }
  #else
  { BLASNAME(strsv)( UPLO, TRANS, DIAG, &N, A, &LDA, X, &INCX ) ; }
  #endif

  inline
  void
  trsv( character        const UPLO[],
        character        const TRANS[],
        character        const DIAG[],
        integer          const N, 
        double_precision const A[], 
        integer          const LDA, 
        double_precision       X[], 
        integer          const INCX )
  #ifdef USE_ATLAS
  { ATLASNAME(dtrsv)( uplo_atlas(UPLO[0]),
                      trans_atlas(TRANS[0]),
                      diag_atlas(DIAG[0]),
                      N, A, LDA, X, INCX ) ; }
  #else
  { BLASNAME(dtrsv)( UPLO, TRANS, DIAG, &N, A, &LDA, X, &INCX ) ; }
  #endif

  /*
  //   _                            
  //  | |_ _ __ _ __ ___  _ __ ___  
  //  | __| '__| '_ ` _ \| '_ ` _ \ 
  //  | |_| |  | | | | | | | | | | |
  //   \__|_|  |_| |_| |_|_| |_| |_|
  */                              
  extern "C" {
    void
    BLASNAME(strmm) ( char             const   SIDE[],     // "L" or "R" 
                      char             const   UPLO[],     // "U" or "L"
                      char             const   TRANSA[],   // "N", "T", "C"
                      char             const   DIAG[],     // "N", "U"
                      integer          const * M,
                      integer          const * N,
                      single_precision const * ALPHA,
                      single_precision const   A[],
                      integer          const * LDA,
                      single_precision         B[],
                      integer          const * LDB ) ;
    void
    BLASNAME(dtrmm) ( char             const   SIDE[],     // "L" or "R" 
                      char             const   UPLO[],     // "U" or "L"
                      char             const   TRANSA[],   // "N", "T", "C"
                      char             const   DIAG[],     // "N", "U"
                      integer          const * M,
                      integer          const * N,
                      double_precision const * ALPHA,
                      double_precision const   A[],
                      integer          const * LDA,
                      double_precision         B[],
                      integer          const * LDB ) ;

    void
    ATLASNAME(strmm)( const enum ATLAS_SIDE  Side,
                      const enum ATLAS_UPLO  Uplo,
                      const enum ATLAS_TRANS TransA,
                      const enum ATLAS_DIAG  Diag,
                      const int M, 
                      const int N,
                      const float alpha,
                      const float *A, const int lda,
                            float *B, const int ldb ) ;
    void
    ATLASNAME(dtrmm)( const enum ATLAS_SIDE  Side,
                      const enum ATLAS_UPLO  Uplo,
                      const enum ATLAS_TRANS TransA,
                      const enum ATLAS_DIAG  Diag,
                      const int M, 
                      const int N,
                      const double alpha,
                      const double *A, const int lda,
                            double *B, const int ldb ) ;
  }
  
  inline
  void
  trmm( character        const SIDE[],
        character        const UPLO[],
        character        const TRANSA[],
        character        const DIAG[],
        integer          const M,
        integer          const N,
        double_precision const alpha, 
        double_precision const A[], 
        integer          const LDA, 
        double_precision       B[], 
        integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(dtrmm)( side_atlas(SIDE[0]),
                      uplo_atlas(UPLO[0]),
                      trans_atlas(TRANSA[0]),
                      diag_atlas(DIAG[0]),
                      M, N, alpha, A, LDA, B, LDB ) ; }
  #else
  { BLASNAME(dtrmm) ( SIDE, UPLO, TRANSA, DIAG, &M, &N, &alpha, A, &LDA, B, &LDB ) ; }
  #endif

  /*
  //   _                      
  //  | |_ _ __ ___ _ __ ___  
  //  | __| '__/ __| '_ ` _ \ 
  //  | |_| |  \__ \ | | | | |
  //   \__|_|  |___/_| |_| |_|
  */                        
  extern "C" {
      void
    BLASNAME(strsm)( character        const   SIDE[], 
                     character        const   UPLO[], 
                     character        const   TRANSA[], 
                     character        const   DIAG[], 
                     integer          const * M,
                     integer          const * N,
                     single_precision const * alpha, 
                     single_precision const   A[], 
                     integer          const * LDA, 
                     single_precision         B[], 
                     integer          const * LDB ) ;
    void
    BLASNAME(dtrsm)( character        const   SIDE[], 
                     character        const   UPLO[], 
                     character        const   TRANSA[], 
                     character        const   DIAG[], 
                     integer          const * M,
                     integer          const * N, 
                     double_precision const * alpha, 
                     double_precision const   A[], 
                     integer          const * LDA, 
                     double_precision         B[], 
                     integer          const * LDB ) ;
    void
    ATLASNAME(strsm)( const enum ATLAS_SIDE  Side,
                      const enum ATLAS_UPLO  Uplo,
                      const enum ATLAS_TRANS TransA,
                      const enum ATLAS_DIAG  Diag,
                      const int M,
                      const int N,
                      const float alpha,
                      const float *A, const int lda,
                            float *B, const int ldb ) ;
    void
    ATLASNAME(dtrsm)( const enum ATLAS_SIDE  Side,
                      const enum ATLAS_UPLO  Uplo,
                      const enum ATLAS_TRANS TransA,
                      const enum ATLAS_DIAG  Diag,
                      const int M,
                      const int N,
                      const double alpha,
                      const double *A, const int lda,
                            double *B, const int ldb ) ;
  }

  inline
  void
  trsm( character        const SIDE[], 
        character        const UPLO[], 
        character        const TRANS[], 
        character        const DIAG[],
        integer          const M,
        integer          const N,
        single_precision const alpha, 
        single_precision const A[], 
        integer          const LDA, 
        single_precision       B[], 
        integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(strsm)( side_atlas(SIDE[0]),
                      uplo_atlas(UPLO[0]),
                      trans_atlas(TRANS[0]),
                      diag_atlas(DIAG[0]),
                      M, N, alpha, A, LDA, B, LDB ) ; }
  #else
  { BLASNAME(strsm)( SIDE, UPLO, TRANS, DIAG, &M, &N, &alpha, A, &LDA, B, &LDB ) ; }
  #endif

  inline
  void
  trsm( character        const SIDE[], 
        character        const UPLO[], 
        character        const TRANS[], 
        character        const DIAG[],
        integer          const M,
        integer          const N,
        double_precision const alpha, 
        double_precision const A[], 
        integer          const LDA, 
        double_precision       B[], 
        integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(dtrsm)( side_atlas(SIDE[0]),
                      uplo_atlas(UPLO[0]),
                      trans_atlas(TRANS[0]),
                      diag_atlas(DIAG[0]),
                      M, N, alpha, A, LDA, B, LDB ) ; }
  #else
  { BLASNAME(dtrsm)( SIDE, UPLO, TRANS, DIAG, &M, &N, &alpha, A, &LDA, B, &LDB ) ; }
  #endif

} // end namespace alglin

#endif

///
/// eof: alglin.hh
///

